"""
    Simple neural network code to load/save/process NN models and training data

    M. Doucet
"""
import os
import yaml
import time
import json

import numpy as np
from sklearn import preprocessing
from tensorflow import keras
from keras.models import model_from_json
from keras.utils import CustomObjectScope
from keras.initializers import glorot_uniform

from lhs_generator import lhs
import factory


def save_model(model, model_name, data_dir=''):
    """
        Save a trained model
        @param model: TensorFlow model
        @param model_name: base name for saved files
        @param data_dir: output directory
    """
    # serialize model to JSON
    model_json = model.to_json()
    with open(os.path.join(data_dir, "%s.json" % model_name), "w") as json_file:
        json_file.write(model_json)
    # serialize weights to HDF5
    model.save_weights(os.path.join(data_dir, "%s.h5" % model_name))


def load_model(model_name, data_dir=''):
    """
        Load and return a trained model
        @param model_name: base name for saved files
        @param data_dir: directory containing trained model
    """
    # load json and create model
    json_file = open(os.path.join(data_dir, '%s.json' % model_name), 'r')
    loaded_model_json = json_file.read()
    json_file.close()

    with CustomObjectScope({'GlorotUniform': glorot_uniform()}):
        model = model_from_json(loaded_model_json)

    # load weights into new model
    model.load_weights(os.path.join(data_dir, '%s.h5' % model_name))
    return model


class TrainingSet:
    """
        Utility class to perform operations on a training set
    """
    validation_data = None
    validation_pars = None
    train_data = None
    train_pars = None
    do_preprocess = False
    scaler_pars = []

    def __init__(self, model_file=None, data_file=None):
        """
            @param model_file: XML file describing the parameters of our thin film model
            @param data_file: test data file (real data) used to pick Q binning
        """
        self.model_config = model_file
        self.data_file = data_file
        # Create q-values as needed
        if data_file is None:
            self.q = np.logspace(np.log10(0.005), np.log10(0.2), num=80)
        else:
            self.q = self.q_values_from_file(data_file)

    def _get_config_name(self):
        """
            Return a model configuration name
        """
        _, _name = os.path.split(self.model_config)
        _config_name, _ = os.path.splitext(_name)
        return _config_name

    def save(self, output_dir=''):
        """
            Save all data relevant to a training set
            @param output_dir: directory used to store training sets
        """
        _config_name = self._get_config_name()

        # Save q values
        np.save(os.path.join(output_dir, "%s_q_values" % _config_name), self.q)
        if self.do_preprocess:
            np.save(os.path.join(output_dir, "%s_scaler_pars" % _config_name), self.scaler_pars)

        # Save parameter names
        with open(os.path.join(output_dir, "%s_par_names.json" % _config_name), 'w') as fd:
            json.dump(self.pars, fd)

        # Save training set
        if self.train_data is not None:
            np.save(os.path.join(output_dir, "%s_data" % _config_name), self.train_data)
            np.save(os.path.join(output_dir, "%s_pars" % _config_name), self.train_pars)

        # Save validation set
        if self.validation_data is not None:
            np.save(os.path.join(output_dir, "%s_validation_data" % _config_name), self.validation_data)
            np.save(os.path.join(output_dir, "%s_validation_pars" % _config_name), self.validation_pars)

    def load(self, data_dir=''):
        _config_name = self._get_config_name()
        print("Loading: %s" % _config_name)
        with open(os.path.join(data_dir, "%s_par_names.json" % _config_name), 'r') as fd:
            self.pars = json.load(fd)
        self.q = np.load(os.path.join(data_dir, "%s_q_values.npy" % _config_name))
        self.train_data = np.load(os.path.join(data_dir, "%s_data.npy" % _config_name))
        self.train_pars = np.load(os.path.join(data_dir, "%s_pars.npy" % _config_name))
        self.validation_data = np.load(os.path.join(data_dir, "%s_validation_data.npy" % _config_name))
        self.validation_pars = np.load(os.path.join(data_dir, "%s_validation_pars.npy" % _config_name))
        return self.q, self.train_data, self.train_pars, self.validation_data, self.validation_pars

    def q_values_from_file(self, data_file):
        data_off = np.loadtxt(data_file).T
        above_zero = data_off[1]>0

        self.q = data_off[0][above_zero]
        self.iq_ref = data_off[1][above_zero]
        if len(data_off) > 2:
            self.diq_ref = data_off[2][above_zero]
        else:
            self.diq_ref = np.zeros(len(self.q))
        return self.q

    def reverse(self, pars):
        return pars
        _pars = 1.0 * pars
        _pars[:, 1] = pars[:, 1] * 500.0
        _pars[:, 4] = pars[:, 4] * 500.0
        return _pars

    def preprocess(self):
        """ Pre-process data """
        self.train_data = -np.log10(self.train_data)
        return self.train_data

    def preprocess_data(self, add_noise=False, sigma=None):
        """ Pre-process data """
        if add_noise:
            if sigma is None:
                _sigma = self.diq_ref
            else:
                _sigma = np.ones(len(self.iq_ref)) * sigma * self.iq_ref
            data_ = self.iq_ref + np.random.normal(size=len(self.iq_ref)) * _sigma
            data_ = np.fabs(data_)
        else:
            data_ = self.iq_ref
        self.iq_ref_proc = -np.log10(data_ * self.q**4)
        return self.iq_ref_proc

    def split(self, n_split=100):
        if n_split < self.train_data.shape[0]:
            self.validation_data = self.train_data[:100]
            self.validation_pars = self.train_pars[:100]

            self.train_data = self.train_data[100:]
            self.train_pars = self.train_pars[100:]
        else:
            raise RuntimeError("Not enough data for splitting %g < %g", self.train_data.shape[0], n_split)
        return self.train_data, self.train_pars, self.validation_data, self.validation_pars

    def _process_factory(self, model_file=None):
        if model_file is not None:
            self.model_config = model_file
        q = self.q
        class_info = yaml.load(open(self.model_config, 'r'))#, Loader=yaml.FullLoader)
        n_classes = len(class_info.keys())
        _factory = factory.new(self.model_config)
        parm_dic = {}
        name = list(class_info.keys())[0]
        return _factory.factories[str(name)].generate_one()

    def evaluate_model(self, params, values, model_file=None, set_ref=False):

        model = self._process_factory(model_file=model_file)
        for i, par in enumerate(params):
            model.setParam(par, values[i])
        iq_model, z_model, sld_model = model.evalDistribution(self.q, True)

        if set_ref:
            self.iq_ref = iq_model
            self.diq_ref = np.zeros(len(self.q))

        return iq_model, z_model, sld_model

    def probability(self, iq, sigma=None):
        """
            Compute chi2 and assign a probability


            TODO: if sigma is not None
            assign errors when computing chi2
        """
        errors = self.diq_ref
        if np.sum(self.diq_ref) == 0:
            errors = np.ones(len(self.q))
        if sigma is not None:
            errors = np.ones(len(self.iq_ref)) * sigma * self.iq_ref
        logp = 0.5 * np.sum((iq - self.iq_ref)**2/errors**2) / len(iq)
        return logp

    def generate_training_set(self, model_file=None, n_samples=1000, error=False, sigma=0.05):
        """
            Uses KNN to guess best parameters from the models given in the yaml file
            :param q: Points that data is sampled.
            :param data: Intensity values for each point
            :param model_file: Yaml file (full path).
            :param n_samples: Number of samples per class
            :param verbose: Display algorithmic progress
        """
        if model_file is not None:
            self.model_config = model_file
        q = self.q
        class_info = yaml.load(open(self.model_config, 'r'))#, Loader=yaml.FullLoader)
        n_classes = len(class_info.keys())
        _factory = factory.new(self.model_config)
        parm_dic = {}

        print(list(class_info.keys()))

        name = list(class_info.keys())[0]

        model = _factory.factories[str(name)].generate_one()
        parm_range = np.zeros([2,0])
        parm_names = []
        for pname in model.params:
            prop_pname = class_info[name][pname].keys()
            if 'linear' in prop_pname:
                parm_names.append(pname)
                int_range = np.zeros([2,1])
                int_range[0,0] = float(class_info[name][pname]['linear'][0])
                int_range[1,0] = float(class_info[name][pname]['linear'][1])
                parm_range = np.append(parm_range,int_range, axis=1)

            if 'log' in prop_pname:
                parm_names.append(pname)
                parm_range = np.append(parm_range,np.reshape(np.array(class_info[name][pname]['linear']),[2,1]), axis=1)

        nparm = parm_range.shape[1]

        t_0 = time.time()
        if n_samples < 1000:
            n_per_slice = n_samples
        else:
            n_per_slice = 1000

        n_slice = np.int(np.floor(n_samples/n_per_slice))
        n_samples = n_slice * n_per_slice
        parm = np.zeros([n_samples, nparm])

        for i_slice in range(n_slice):
            _parm = lhs(nparm, samples=n_per_slice, iterations=5)
            parm[i_slice*n_per_slice:(i_slice+1)*n_per_slice, :] = _parm

        print("Training set generated in %s sec" % (time.time()-t_0))

        for jp in range(nparm):
            parm[:,jp] = (parm_range[1,jp]-parm_range[0,jp])*parm[:,jp]+parm_range[0,jp]
            prop_pname = class_info[name][parm_names[0]].keys()
            if 'log' in prop_pname:
                parm[:,jp] = np.exp(parm[:,jp])

        data = []
        parm_ = []
        for js in range(n_samples):
            for jp in range(nparm):
                model.setParam(parm_names[jp],parm[js,jp])
            data_, _, _ = model.evalDistribution(q)
            if error:
                data_ *= (1 + np.random.normal(size=len(q))*sigma)
            data_ *= q**4
            data.append(data_)
            parm_.append(parm[js, :])

        self.pars = parm_names
        self.train_data = np.asarray(data)
        self.train_pars = np.asarray(parm_)
        return self.pars, self.train_data, self.train_pars

    def process_pars_file(self, error=False, sigma=0.05):

        nparm = len(self.pars)
        n_samples = self.train_pars.shape[0]
        print("n_samples %s" % n_samples)

        model = self._process_factory(model_file=self.model_config)

        data = []
        for js in range(n_samples):
            for jp in range(nparm):
                model.setParam(self.pars[jp], self.train_pars[js,jp])
            data_, _, _ = model.evalDistribution(self.q)
            if error:
                data_ *= (1 + np.random.normal(size=len(q))*sigma)
            data_ *= self.q**4
            data.append(data_)

        self.train_data = np.asarray(data)
        return self.pars, self.train_data, self.train_pars
