"""
    Code recycled as-is from LDRD project
"""
import numpy as np
import time as t
import math as m
from .import config_parser
from .reflectivity_model import ReflectivityModel
from pkg_resources import resource_filename


MODEL_CONFIG = resource_filename('factory','ml_train_config_short.yaml')


_global_rng = None
def get_rng():
    """ This is a single-threaded application, so we only need one global
        random number generator """
    global _global_rng
    if not _global_rng:
        np.random.seed(int(t.time()))
        _global_rng = np.random.random
    return _global_rng


Infinity = float("inf")


def new(config_file=MODEL_CONFIG, max_model=Infinity):
    config = config_parser.load(config_file)
    factories = FactoryManager(max_model)
    for factory_name, factory_config in config.items():
        factory = new_factory(factory_config)
        factories.add_factory(factory_name, factory)
    return factories


def add_range_to_factory(factory, parameter, value):
    if 'fixed' in value:
        factory.add_fixed(parameter, value['fixed'][0])
    elif 'linear' in value:
        factory.add_range(parameter, value['linear'][0], value['linear'][1])
    elif 'log' in value:
        factory.add_log(parameter, value['log'][0], value['log'][1])
    elif 'integer' in value:
        factory.add_range(parameter, value['integer'][0], value['integer'][1])

    if 'coupled' == value:
        target_parameter = value['coupled'][0]
        factory.add_coupling_dependency(
                dependent=parameter, independent=target_parameter)


def new_factory(shape_definition):
    """ Build a new ModelFactory from a shape definition in the shape_definition
        file """
    f = ModelFactory(shape_definition['model_name'])
    del shape_definition['model_name']
    del shape_definition['Category']  # Not *yet* doing anything with this
    if 'Model-Type' in shape_definition:
        f.model_type = shape_definition['Model-Type']

    # Set Q range as appropriate
    if 'q_range' in shape_definition:
        f.q_range = shape_definition['q_range']
    for parameter in shape_definition:
        value = shape_definition[parameter]
        # pick one of ['fixed', 'linear', 'log', 'integer']
        add_range_to_factory(f, parameter, value)

    return f


class FactoryManager(object):
    """ Coordinates a collection of Model Factories. Provides an iterator
        which supplies a stream of TheoryCurve objects, each of which is
        constructed by choosing a factory at random, and then sampling
        that factory """
    def __init__(self, max_model):
        self.factories = {}
        self.max_model = max_model
        self.current_model = 0
        self.rng = get_rng()

    def add_factory(self, name, factory):
        """ Declare a new model factory """
        self.factories[name] = factory

    def __iter__(self):
        return self

    def next(self):
        self.current_model += 1
        if self.current_model < self.max_model:
            return self.random_model()
        else:
            raise StopIteration

    def random_model(self):
        """ Selects a factory at random, then samples a model from that
            factory and returns a TheoryCurve object based on that model """
        factory = self._random_factory()
        model = factory.generate_one()
        return TheoryCurve(model)

    def _random_factory(self):
        key = self._random_factory_key()
        return self.factories[key]

    def _random_factory_key(self):
        index = int(self._num_factories * self.rng())
        return self._factory_keys[index]

    @property
    def _factory_keys(self):
        return self.factories.keys()

    @property
    def _num_factories(self):
        return len(self._factory_keys)


class Range(object):
    """ Represents a random variable with a uniform distribution over
        some asserted range """
    def __init__(self, lower_bound, upper_bound):
        #assert float(lower_bound) <= float(upper_bound)
        self.lower_bound = float(lower_bound)
        self.upper_bound = float(upper_bound)
        self.rng = get_rng()

    @property
    def width(self):
        """ The distance between the lower bound and upper bound """
        return self.upper_bound - self.lower_bound

    def random(self):
        """ Sample a value from the range """
        return self.width * self.rng() + self.lower_bound


class IntegerRange(Range):
    """ Random Variable over Z """
    def random(self):
        """ Select an integer value for the sampled variable """
        return int(super(IntegerRange,self).random())


class LogRange(Range):
    """ Random Variable on a log-scale """
    def random(self):
        """ Pick uniformly from a range of exponents, then return 10 raised
            to that power """
        return float(10 ** super(LogRange,self).random())


class ModelFactory(object):
    """ Houses a collection of random variables from which a model instance
        can be sampled at runtime. """
    def __init__(self, model_name):
        self.model_name = model_name
        self.ranges = {}
        self.fixed_values = {}
        self.dispersion = {}
        self.coupling_dependencies = {}
        self.cached_values = {}
        self.model_type = 'sasview'

    def add_log(self, name, lower, upper):
        """ Declare a Log-scale Range """
        self.ranges[name] = LogRange(lower, upper)

    def add_range(self, name, lower, upper):
        """ Declare a linear range """
        self.ranges[name] = Range(lower, upper)

    def add_integer_range(self, name, lower, upper):
        """ Declare an integer range """
        self.ranges[name] = IntegerRange(lower, upper)

    def add_fixed(self, name, value):
        """ add a fixed parameter """
        self.fixed_values[name] = float(value)

    def add_coupling_dependency(self, dependent=None, independent=None):
        self.coupling_dependencies[dependent] = independent

    def generate_one(self):
        """ Prepare a sasmodels object from the"""
        model = ReflectivityModel(self.model_name)
        for range_name, range_value in self.ranges.items():
            self.cached_values[range_name] = range_value.random()
        for fixed_name, fixed_value in self.fixed_values.items():
            self.cached_values[fixed_name] = fixed_value
        for dependent, independent in self.coupling_dependencies.items():
            self.cached_values[dependent] *= self.cached_values[independent]
        for name, value in self.cached_values.items():
            model.setParam(name, value)
        return model


COMMON_Q_RANGE = np.logspace(-3.0, 0.0, num=1000, endpoint=True, base=10, dtype=None)

class TheoryCurve(object):
    """ Represents a fixed domain q and the range I(q) """
    def __init__(self, model):
        self.model = model
        if hasattr(model, 'q_range') and model.q_range is not None:
            self.q = np.logspace(np.log(model.q_range[0]), np.log(model.q_range[1]), num=1000, endpoint=True, base=10, dtype=None)
        else:
            self.q = COMMON_Q_RANGE

    def iq(self):
        """ Intensity Function (Calculated on the fly) """
        return self.model.evalDistribution(self.q)

    @property
    def model_name(self):
        """ Name of model to be used for classification """
        return "%s" % self.model

    def __str__(self):
        """ String representation to be used for debugging / inspection """
        return "%s | %s" % (self.model, self.iq())
