"""
    Generate a reflectivity model

    Refl1d must be installed:
    > pip install refl1d

    This code was recycled from a previous LDRD project
"""
from __future__ import absolute_import, division, print_function, unicode_literals
import collections
import numpy as np


from refl1d.names import *


class ReflectivityModel(object):
    """
        Refl1d model in the style of sasmodels
    """
    def __init__(self, model_name="sample"):
        self.model_name = model_name
        self.params = collections.OrderedDict()
        self.params['n_layers'] = 0
        self.params['background'] = 0
        self.params['front_sld'] = 0
        self.params['back_sld'] = 2.07
        self.params['back_roughness'] = 5
        self.q_range = None

    def setParam(self, name, value):
        """
            Set the value of a model parameter
            @param name: name of the parameter
            @param value: value of the parameter

        """
        self.params[name] = value

    def getParam(self, name):
        """
            Return the value of a parameter
            @param name: name of the parameter
        """
        if name in self.params.keys():
            return self.params[name]
        else:
            raise ValueError("Model does not contain parameter %s" % name)

    def verify_parameters(self):
        """
            Verify the integrity of the model
        """
        if 'n_layers' not in self.params:
            raise RuntimeError("Number of layers not specified")
        if 'front_sld' not in self.params:
            raise RuntimeError("front_sld parameter not specified")
        if 'back_sld' not in self.params:
            raise RuntimeError("back_sld parameter not specified")
        if 'back_roughness' not in self.params:
            raise RuntimeError("back_roughness parameter not specified")
        if 'background' not in self.params:
            raise RuntimeError("background parameter not specified")
        for i in range(int(self.params['n_layers'])):
            if 'sld-%d' % i not in self.params:
                raise RuntimeError("Layer %s is missing an SLD", i)
            if 'thickness-%d' % i not in self.params:
                raise RuntimeError("Layer %s is missing a thickness", i)
            if 'roughness-%d' % i not in self.params:
                raise RuntimeError("Layer %s is missing a roughness", i)

    def evalDistribution(self, qdist, do_sld=False):
        """
            Calculate reflectivity. Returns a numpy ndarray.
            @param qdist: ndarray of q-values
        """
        self.verify_parameters()
        if isinstance(qdist, np.ndarray):
            # We have a simple 1D distribution of q-values
            return self.calculate_reflectivity(qdist, do_sld)

        else:
            raise TypeError("evalDistribution expects q, not %r" % type(qdist))

    def calculate_reflectivity(self, q, do_sld=False, amplitude=False):
        """
            Reflectivity calculation using refl1d
        """
        # Q array
        zeros = np.zeros(len(q))
        # this is new 4/24/2019 to smear the profile - William
        dq = 0.025*q
        #probe = QProbe(q, zeros, data=(zeros, zeros))
        # this is new 4/24/2019 to smear the profile - William
        probe = QProbe(q, dq, data=(zeros, zeros))

        sample = Slab(material=SLD(name='back',
                                         rho=self.params['back_sld']),
                         interface=self.params['back_roughness'])
        for i in range(int(self.params['n_layers'])):
            sample = sample | Slab(material=SLD(name='layer%s' % i,
                                                  rho=self.params['sld-%d' % i],
                                                  irho=0),
                                      thickness=self.params['thickness-%d' % i],
                                      interface=self.params['roughness-%d' % i])

        sample = sample | Slab(material=SLD(name='front',
                                                  rho=self.params['front_sld']))

        #probe.intensity = rf.Parameter(value=fit_problem.reflectivity_model.scale, name='scale')
        probe.background=Parameter(value=self.params['background'],name='background')
        expt = Experiment(probe=probe, sample=sample)
        if amplitude:
            q, r = expt.amplitude()
        else:
            q, r = expt.reflectivity()

        if hasattr(self, '_z_left'):
            expt._slabs._z_left = self._z_left
        if hasattr(self, '_z_right'):
            expt._slabs._z_right = self._z_right
        if do_sld:
            z, sld, _ = expt.smooth_profile()
        else:
            z = None
            sld = None
        return r, z, sld
